-- ============================================================================
-- == HTCommon - A collection of common functions used throughout the        ==
-- ==            HalonaTools addon.                                          ==
-- ==                                                                        ==
-- ==            This includes message text display... TBC.                  ==
-- ============================================================================

HTCommon           = LibStub("AceAddon-3.0"):NewAddon("HTCommon", "AceConsole-3.0", "AceEvent-3.0")
local L            = LibStub("AceLocale-3.0"):GetLocale("HalonaTools")

local officerRanks = {}           -- Indicates which ranks are 'officer' ranks.

-- ============================================================================
-- == OnInitialize()                                                         ==
-- ==                                                                        ==
-- == This function is called when the addon is initialised.                 ==
-- ============================================================================

function HTCommon:OnInitialize()
end

-- ============================================================================
-- == OnEnable()                                                             ==
-- ==                                                                        ==
-- == This function is called when the addon is enabled.                     ==
-- ============================================================================

function HTCommon:OnEnable()

  HTCommon:ConfigureOfficerRanks()

end

-- ============================================================================
-- == OnEnable()                                                             ==
-- ==                                                                        ==
-- == This function is called when the addon is disabled.                    ==
-- ============================================================================

function HTCommon:OnDisable()
end

-- ============================================================================
-- == GetGuildName()                                                         ==
-- ==                                                                        ==
-- == This function retrieves the Guild name of the user, if they are in a   ==
-- == guild, otherwise it returns "Unguilded".                               ==
-- ============================================================================

function HTCommon:GetGuildName()

  if IsInGuild() then
    local guildName, guildRankName, guildRankIndex = GetGuildInfo("player")
    return guildName
  else
    return L["common-guildname-unguilded"]
  end

end

-- ============================================================================
-- == GetGuildRankName()                                                     ==
-- ==                                                                        ==
-- == This function retrieves the rank name of the user, if they are in a    ==
-- == guild, otherwise it returns "Unranked".                                ==
-- ============================================================================

function HTCommon:GetGuildRankName()

  if IsInGuild() then
    local guildName, guildRankName, guildRankIndex = GetGuildInfo("player")
    return guildRankName
  else
    return L["common-rankname-unguilded"]
  end

end

-- ============================================================================
-- ==  GetGuildLevel()                                                       ==
-- ==                                                                        ==
-- == This function retrieves the level of the player's guild.               ==
-- ============================================================================

function HTCommon:GetGuildLevel()

  if IsInGuild() then
    return GetGuildLevel()
  else
    return L["common-rankname-unguilded"]
  end

end

-- ============================================================================
-- == GetGuildOfficerList()                                                  ==
-- ==                                                                        ==
-- == This function retrieves a list of all main-rank "officer" rated        ==
-- == characters in the guild.                                               ==
-- ============================================================================

function HTCommon:GetGuildOfficerList()

  local guildOfficerList = {}

  -- HTConsole:Debug("HTCommon:GetGuildOfficerList() has been called.")

  for index = 1, (GetNumGuildMembers(true)) do
    local charName, charRank, charRankIndex, charLevel, charClass, charZone, charNote, charOfficernote, charOnline, charStatus = GetGuildRosterInfo(index);

    -- HTConsole:Debug("Evaluating character: [" .. charName .. "] with rank: [" .. charRank .. ", rankIndex: [" .. charRankIndex .. "]")

    local rankIndex = charRankIndex + 1

    -- HTConsole:Debug("officerRanks[" .. rankIndex .. "] = [" .. officerRanks[rankIndex] .. "]")
    -- HTConsole:Debug("SafeNumber = [" .. HTData:SafeNumber(officerRanks[rankIndex])  .. "]")

    if HTData:SafeNumber(officerRanks[rankIndex]) == "1" and HTCommon:IsRankIndexTaggedAs(rankIndex, "main") then
      table.insert(guildOfficerList, charName)
      -- HTConsole:Debug("- this character IS an officer and a main!")
    end

  end

  return guildOfficerList

end

-- ============================================================================
-- == FindPlayerByCharacterName(mainCharacterName)                           ==
-- ==                                                                        ==
-- == This function takes a character name as input and searchs through the  ==
-- == guild's roster, looking to find if the player that owns that character ==
-- == is currently online, either on the character named or on an alt        ==
-- == character associated with the character named.                         ==
-- ==                                                                        ==
-- == This assumes that for a given main character, all associated alt       ==
-- == characters will have the main character's name in the public note      ==
-- == field.                                                                 ==
-- ============================================================================

function HTCommon:FindPlayerByCharacterName(characterName)

  characterNameToReturn = characterName .. L["con-findbycharname-offline"]

  totalNumberOfGuildMembers = GetNumGuildMembers(true)

  for index=1, totalNumberOfGuildMembers, 1 do
    name, rank, rankIndex, level, class, zone, note, officernote, online, status = GetGuildRosterInfo(index)
    if name == characterName and online == 1 then
      characterNameToReturn = characterName
    elseif note == characterName and online == 1 then
      characterNameToReturn = name .. L["con-findbycharname-aka"] .. characterName .. ")"
    end
  end

  return characterNameToReturn

end

-- ============================================================================
-- == GetSimpleCharacterName(complexCharacterName)                           ==
-- ==                                                                        ==
-- == This function takes a string, typically assumed to be the output of    ==
-- == FindPlayerByCharacterName(mainCharacterName) function and attempts to  ==
-- == simplify it back from the complex form it may hold to the more simple  ==
-- == form.                                                                  ==
-- ==                                                                        ==
-- == Examples of complexCharacterName conversions.                          ==
-- ==                                                                        ==
-- == Broliant (Offline)         -> Broliant                                 ==
-- == Balgardner (aka: Broliant) -> Balgardner                               ==
-- == Broliant                   -> Broliant                                 ==
-- ==                                                                        ==
-- ============================================================================

function HTCommon:GetSimpleCharacterName(complexCharacterName)

    if HTConfig:GetIsEnabledDebugMode(nil) then
       -- HTConsole:Debug("HTCommon:GetSimpleCharacterName("..complexCharacterName..")")
    end

    if not string.match(complexCharacterName, " ") == nil then
       local Args = HTData:StringSplit(complexCharacterName, " ")
       -- HTConsole:Debug("Returning: "..Args[1]..".")
       return Args[1]
    else
       -- HTConsole:Debug("Returning: "..complexCharacterName..".")
       return complexCharacterName
    end

end

-- ============================================================================
-- == IsRankIndexTaggedAs(rankIndex, tag)                                    ==
-- ==                                                                        ==
-- == This function determines whether the supplied rankIndex value          ==
-- == corresponds to the supplied tag value and returns true if it does and  ==
-- == false otherwise.                                                       ==
-- ==                                                                        ==
-- == Examples:                                                              ==
-- == Is a rankIndex a main rank? : IsRankIndexTaggedAs(rankIndex, "main")   ==
-- == Is a rankIndex a trial rank?: IsRankIndexTaggedAs(rankIndex, "trial")  ==
-- ============================================================================

function HTCommon:IsRankIndexTaggedAs(rankIndex, tag)

  if L["common-guildname-unguilded"] == HTCommon:GetGuildName() then
    return false
  end

  local guildName = HTData:AsResourceName(string.lower(HTCommon:GetGuildName()))
  local resourceID = guildName .. "-rank-" .. HTData:PadLeadingZeros(rankIndex+1, 2) .. "-tags"
  local rankTags = L[resourceID]

  if HTCommon:IsTaggedWith(rankTags, tag) then
    return true
  else
    return false
  end

end

-- ============================================================================
-- == IsTaggedWith(text, tag)                                                ==
-- ==                                                                        ==
-- == This function determines whether the supplied text contains the named  ==
-- == tag.                                                                   ==
-- ============================================================================

function HTCommon:IsTaggedWith(text, tag)

  local tagList = HTData:StringSplit(text, ",")
  for i,v in ipairs(tagList) do

    if v == tag then
      return true
    end

  end

  return false

end

-- ============================================================================
-- ==  ConfigureOfficerRanks()                                               ==
-- ==                                                                        ==
-- == This function configures the 'officerRanks' data table to indicate     ==
-- == which of the ten possible guild ranks currently in use are considered  ==
-- == to be 'Officer Ranks'. This is done by assuming that an 'officer' rank ==
-- == is one which has been granted the privilege of allowing the user the   ==
-- == ability to 'speak' in the officer chat channel.                        ==
-- ==                                                                        ==
-- == The resulting 'officerRanks' table will be a 1 dimensional array       ==
-- == containing 10 indexes where officerRanks[1] corresponds to guild rank  ==
-- == 1 (the 'GM' rank) and officerRanks[10] corresponds to guild rank 10.   ==
-- == If a guild rank is an 'officer' rank then the index holds the value of ==
-- == '1', otherwise the index holds the value of '0' indicating the rank is ==
-- == either not an 'officer' rank or not in use.                            ==
-- ==                                                                        ==
-- == It is expected that this function is called when the addon is loaded   ==
-- == or initialised.                                                        ==
-- ============================================================================

function HTCommon:ConfigureOfficerRanks()

  for i=1,10 do

    GuildControlSetRank(i)
    local guildchat_listen, guildchat_speak, officerchat_listen, officerchat_speak, promote, demote, invite_member, remove_member, set_motd, edit_public_note, view_officer_note, edit_officer_note, modify_guild_info, withdraw_repair, withdraw_gold, create_guild_event = GuildControlGetRankFlags()
    table.insert(officerRanks, HTData:SafeNumericBoolean(officerchat_speak))
    -- HTConsole:Debug("i=" .. i ..", officer-rank-flag=" .. HTData:SafeNumericBoolean(officerchat_speak))

  end

end

-- ============================================================================
-- ==  IsOfficerRank(rankIndex)                                              ==
-- ==                                                                        ==
-- == This function determines whether the supplied rankIndex parameter      ==
-- == represents an "officer-class" guild rank.                              ==
-- ============================================================================

function HTCommon:IsGuildMasterRank(rankIndex)

  if HTData:SafeNumber(rankIndex) == 1 then
    return true
  else
    return false
  end

end

-- ============================================================================
-- ==  IsOfficerRank(rankIndex)                                              ==
-- ==                                                                        ==
-- == This function determines whether the supplied rankIndex parameter      ==
-- == represents an "officer-class" guild rank.                              ==
-- ============================================================================

function HTCommon:IsOfficerRank(rankIndex)

  rankIndex = HTData:SafeNumber(rankIndex)

  local isOfficerRank = false

  if rankIndex < 1 or rankIndex > 10 then
    return isOfficerRank
  end

  GuildControlSetRank(rankIndex)
  local guildchat_listen, guildchat_speak, officerchat_listen, officerchat_speak, promote, demote, invite_member, remove_member, set_motd, edit_public_note, view_officer_note, edit_officer_note, modify_guild_info, withdraw_repair, withdraw_gold, create_guild_event = GuildControlGetRankFlags()

  if HTData:SafeNumericBoolean(officerchat_speak) == 1 then
    isOfficerRank = true
  end

  return isOfficerRank

end

-- ============================================================================
-- ==  IsOfficer(characterName)                                              ==
-- ==                                                                        ==
-- == This function determines whether the character identified by the       ==
-- == supplied name parameter holds an officer level rank within the guild.  ==
-- ============================================================================

function HTCommon:IsOfficer(characterName)

  characterName = HTData:SafeString(characterName)

  for index = 1, (GetNumGuildMembers(true)) do
    local charName, charRank, charRankIndex, charLevel, charClass, charZone, charNote, charOfficernote, charOnline, charStatus = GetGuildRosterInfo(index);

    if charName == characterName then

      -- HTConsole:Debug("CharacterName=" .. charName .. ", officerRanks[" .. charRankIndex .. "+1]=[" .. officerRanks[charRankIndex+1] .. "]")

      if HTData:SafeNumber(officerRanks[charRankIndex+1]) == "1" then
        return true
      end
    end
  end

  return false

end

-- ============================================================================
-- ==  IsCharacterRankedAs(characterName, rankType)                          ==
-- ==                                                                        ==
-- == This function determines whether the rank of the character identified  ==
-- == by the supplied name parameter is accurately described by the supplied ==
-- == rankType parameter. (eg: "main", "alt", "trial" etc) It depends on the ==
-- == existence of accurate rank-tag information for the player's guild.     ==
-- ============================================================================

function HTCommon:IsCharacterRankedAs(characterName, rankType)

  characterName = HTData:SafeString(characterName)

  for index = 1, (GetNumGuildMembers(true)) do
    local charName, charRank, charRankIndex, charLevel, charClass, charZone, charNote, charOfficernote, charOnline, charStatus = GetGuildRosterInfo(index);
    if charName == characterName then
      -- HTConsole:Debug("charName=["..charName.."], charRankIndex=["..charRankIndex.."]")
      if HTCommon:IsRankIndexTaggedAs(charRankIndex, rankType) then
        return true
      end
    end
  end

  return false

end


-- ============================================================================
-- ==  IsMainRankedCharacter(characterName)                                  ==
-- ==                                                                        ==
-- == This function determines whether the character identified by the       ==
-- == supplied name parameter is a main ranked character. It depends on the  ==
-- == existence of accurate rank-tag information for the player's guild.     ==
-- ============================================================================

function HTCommon:IsMainRankedCharacter(characterName)

  HTCommon:IsCharacterRankedAs(characterName, "main")

end

-- ============================================================================
-- ==  IsAltRankedCharacter(characterName)                                   ==
-- ==                                                                        ==
-- == This function determines whether the character identified by the       ==
-- == supplied name parameter is an alt ranked character. It depends on the  ==
-- == existence of accurate rank-tag information for the player's guild.     ==
-- ============================================================================

function HTCommon:IsAltRankedCharacter(characterName)

  HTCommon:IsCharacterRankedAs(characterName, "alt")

end

-- ============================================================================
-- ==  IsTrialRankedCharacter(characterName)                                 ==
-- ==                                                                        ==
-- == This function determines whether the character identified by the       ==
-- == supplied name parameter is a trial ranked character. It depends on the ==
-- == existence of accurate rank-tag information for the player's guild.     ==
-- ============================================================================

function HTCommon:IsTrialRankedCharacter(characterName)

  HTCommon:IsCharacterRankedAs(characterName, "trial")

end

-- ============================================================================
-- ==  IsMe(characterName)                                                   ==
-- ==                                                                        ==
-- == This function determines whether the character identified by the       ==
-- == supplied name parameter is the user of the addon.                      ==
-- ============================================================================

function HTCommon:IsMe(characterName)

  local nameToCheck = HTData:SafeString(characterName)
  local myName      = HTData:SafeString(UnitName("player"))

  if nameToCheck == myName then
    return true
  else
    return false
  end

end

-- ============================================================================
-- == IsGuildMemberOnline(characterName)                                     ==
-- ==                                                                        ==
-- == This function performs a simple check to determine whether the named   ==
-- == guild character is currently online.                                   ==
-- ============================================================================

function HTCommon:IsGuildMemberOnline(characterName)

  totalNumberOfGuildMembers = GetNumGuildMembers(true)

  for index=1, totalNumberOfGuildMembers, 1 do
    name, rank, rankIndex, level, class, zone, note, officernote, online, status = GetGuildRosterInfo(index)
    if name == characterName then

      if online == 1 then
        return true
      else
        return false
      end
    end
  end

end

-- ============================================================================
-- ==  IsCharacterInGuild(characterName)                                     ==
-- ==                                                                        ==
-- == This function determines whether the specified character is in the     ==
-- == player's guild.                                                        ==
-- ============================================================================

function HTCommon:IsCharacterInGuild(characterName)

  -- If the player is not in a guild, no other character can possibly be in the same guild!
  if HTCommon:GetGuildName() == L["common-guildname-unguilded"] then return false end

  -- Check if characterName appears in the guild's character roster.
  for index = 1, (GetNumGuildMembers(true)) do

    local charName, charRank, charRankIndex, charLevel, charClass, charZone, charNote, charOfficernote, charOnline, charStatus = GetGuildRosterInfo(index);

    if charName == characterName then
      return true
    end
  end

  return false

end

-- ============================================================================
-- == IsCharacterInParty(characterName)                                      ==
-- ==                                                                        ==
-- == This function determines whether the named character is in the same    ==
-- == party as the person running HT.                                        ==
-- ============================================================================

function HTCommon:IsCharacterInParty(characterName)

  if UnitInParty(characterName) then
     return true
  end

  return false
end

-- ============================================================================
-- == IsCharacterInRaid(characterName)                                       ==
-- ==                                                                        ==
-- == This function determines whether the named character is in the same    ==
-- == raid as the person running HT.                                         ==
-- ============================================================================

function HTCommon:IsCharacterInRaid(characterName)

  if UnitInRaid(characterName) then
     return true
  end

  return false
end

-- ============================================================================
-- == IsCharacterAFriend(characterName)                                      ==
-- ==                                                                        ==
-- == This function determines whether the named character is a friend of    ==
-- == the person running HT.                                                 ==
-- ============================================================================

function HTCommon:IsCharacterAFriend(characterName)

  local numFriends = GetNumFriends()
  local name, level, class, zone, online, status, note

  for i = 1, numFriends, 1 do
    name, level, class, zone, online, status, note = GetFriendInfo(i)
    if online then
      if name == characterName then
        return true
      end
    end
  end

  return false

end

-- ============================================================================
-- == IsCharacterARealIDFriend(characterName)                                ==
-- ==                                                                        ==
-- == This function determines whether the named character is a friend of    ==
-- == the person running HT.                                                 ==
-- ============================================================================

function HTCommon:IsCharacterARealIDFriend(characterName)

  local numFriendsTotal, numFriendsOnline = BNGetNumFriends()
  local presenceID, givenName, surname, toonName, toonID, client, isOnline, lastOnline, isAFK, isDND, broadcastText, noteText, isFriend, broadcastTime

  for i = 1, numFriendsTotal, 1 do
    presenceID, givenName, surname, toonName, toonID, client, isOnline, lastOnline, isAFK, isDND, broadcastText, noteText, isFriend, broadcastTime = BNGetFriendInfo(i)
    if isOnline then
      if toonName == characterName then
        return true
      end
    end
  end

  return false

end

-- ============================================================================
-- ==  IsPlayerInParty()                                                     ==
-- ==                                                                        ==
-- == This function determines whether the player is currently in a party or ==
-- == not and returns true if the player is in a party, false otherwise.     ==
-- ============================================================================

function HTCommon:IsPlayerInParty()

  local numPartyMembers = HTData:SafeNumber(GetNumPartyMembers())

  if numPartyMembers > 0 then
    return true
  else
    return false
  end

end

-- ============================================================================
-- ==  IsPlayerInRaid()                                                      ==
-- ==                                                                        ==
-- == This function determines whether the player is currently in a raid or  ==
-- == not and returns true if the player is in a raid, false otherwise.      ==
-- ============================================================================

function HTCommon:IsPlayerInRaid()

  local numRaidMembers  = HTData:SafeNumber(GetNumRaidMembers())

  if numRaidMembers > 0 then
    return true
  else
    return false
  end

end
